/*
 *  Digital Instant Camera
 *     for STM32F746 Discovery
 *
 *    Copyright (C) by Y.Ishioka.
 */

#include <string.h>


#define  DEF_PRN_AREA_X   384
#define  DEF_PRN_AREA_Y   (480 + 0)
#define  DEF_PRN_BYTE_Y   (DEF_PRN_AREA_Y/8)

//#define  PRN_PSET_THRED4   280
//#define  PRN_PSET_THRED4   400
#define  PRN_PSET_THRED4   380

unsigned short int  data_surp[272][ 480+1 ] __attribute__ ((section (".usrsec")));
unsigned char  prn_area_data[ DEF_PRN_AREA_X * DEF_PRN_BYTE_Y ] __attribute__ ((section (".usrsec")));

void  prn_send_multibyte( unsigned char *buff, int len ) ;


void  prn_area_clear( void )
{
    memset( prn_area_data, 0x00, sizeof(prn_area_data) ) ;
}


void  prn_area_set( int x, int y )
{
    int  byte ;
    unsigned char  bit ;

    if( x < 0 || x >= DEF_PRN_AREA_X ) {
        return ;
    }

    if( y < 0 || y >= DEF_PRN_AREA_Y ) {
        return ;
    }

    byte = ((y/24) * DEF_PRN_AREA_X * 3)
         + ((y/8) % 3)
         + x * 3 ;

    bit = 0x80 >> (y % 8) ;

    prn_area_data[ byte ] |= bit ;
}


void  prn_area_print( void )
{
    unsigned char   *pnt ;
    unsigned char  *pos ;
    int  len, total ;
    int  i ;
    static  unsigned char  data[10] ;

    len = DEF_PRN_AREA_X * 3 ;
    pos = prn_area_data ;
    total = 0 ;
    for( i=0 ; i<(DEF_PRN_AREA_X * DEF_PRN_BYTE_Y) ; i+=len ) {

        pnt = data ;
        *pnt++ = 0x1b ; /* ESC */
        *pnt++ = 0x2A ; /* * */
        *pnt++ =  33 ;  /* m */
        *pnt++ = DEF_PRN_AREA_X & 0xff ;  /* nL */
        *pnt++ = DEF_PRN_AREA_X >> 8  ;   /* nH */
        (void)prn_send_multibyte( (unsigned char *)data, 5 ) ;

        (void)prn_send_multibyte( (unsigned char *)pos, len ) ;
        pos += len ;
        total += len ;
    }
}


void print_photo_480x272( void )
{
    unsigned char  *pnt ;
    int  len ;
    unsigned char  *src ;
    unsigned short int  cp0, cp1, mono ;
    unsigned short int  surp ;
    int  x, y ;
    static  unsigned char  data[10] ;

    /* printer initialize */
    data[0] = 0x1b ; /* ESC */
    data[1] = 0x40 ; /* @ */
    (void)prn_send_multibyte( (unsigned char *)data, 2 ) ;

    prn_area_clear() ;

    for( x=0 ; x<=480 ; x++ ) {
        for( y=0 ; y<272 ; y++ ) {
            data_surp[y][x] = 0 ;
        }
    }

    /* rbg565 to mono to surp */
    src = (void *)0xc0000000; /* frame top address */
    for( y=0 ; y<272 ; y++ ) {
        for( x=0 ; x<480 ; x++ ) {
            cp1 = (unsigned short int)*src++ ;
            cp0 = (unsigned short int)*src++ ;
            /* (306 * R + 601 * G + 117 * B)/124 */
            mono = ( ((cp0>>3)&0x1f) * 306
                   + (((cp0<<2)&0x1c) + ((cp1>>6)&0x03)) * 601
                   + (cp1 & 0x1f) * 117 ) / 124 ;
            if( mono >= 256 ) {
                mono = 0 ;
            } else {
                mono = 255 - mono ;
            }

            surp = mono + data_surp[y][x] ;
            if( surp >= PRN_PSET_THRED4 ) {
                surp -= PRN_PSET_THRED4 ;
                prn_area_set( y, 480-x ) ;
            }
            data_surp[y  ][x+1] += ((surp * 7) / 16) ;
            data_surp[y+1][x+0] += ((surp * 5) / 16) ;
            data_surp[y+1][x+1] += ((surp    ) / 16) ;
            if( x > 0 ) {
                data_surp[y+1][x-1] += ((surp * 3) / 16) ;
            }
        }
    }

    prn_area_print() ;

    pnt = data ;
    *pnt++ = 0x0a ; /* LF */
    *pnt++ = 0x0a ; /* LF */
    *pnt++ = 0x0a ; /* LF */
    *pnt++ = 0x0a ; /* LF */
    *pnt++ = 0x0c ; /* FF */
    len = pnt - data ;
    prn_send_multibyte( (unsigned char *)data, len ) ;
}


#define  DEF_PRN_PHOT_Y   480

unsigned char  data_buff[ 2 ][ 320*2+1 ] ; /* for rgb565 */
unsigned char  data_exte[ 3 ][ 480+1 ] ;


void print_photo_320x240( void )
{
    unsigned char  *pnt ;
    int  len ;
    int  i ;
    volatile unsigned char  *data_pos ;
    unsigned char  *copy_pos ;
    int  x, y ;
    int  xx, yy ;
    unsigned short int  cp0, cp1, mono ;
    unsigned short int  surp ;
    int  surp_now, surp_next ;
    static  unsigned char  data[10] ;

    /* printer initialize */
    data[0] = 0x1b ; /* ESC */
    data[1] = 0x40 ; /* @ */
    (void)prn_send_multibyte( (unsigned char *)data, 2 ) ;

    prn_area_clear() ;

    for( i=0 ; i<=DEF_PRN_PHOT_Y ; i++ ) {
        data_surp[0][i] = 0 ;
        data_surp[1][i] = 0 ;
    }

    data_pos = (volatile unsigned char*)0x61000002 ;
    /* dummy readd */
    cp0 = *data_pos;
    cp1 = *data_pos;

    yy = 0 ;
    for( y=0 ; y<240 ; y+=2 ) {

        /* rbg565 to mono */
        //copy_pos = data_buff[0] ;
        copy_pos = (void *)(0xc0000000 + (y * 320 * 2)) ;
        for( x=0 ; x<320 ; x++ ) {
            cp1 = (unsigned short int)*copy_pos++ ;
            cp0 = (unsigned short int)*copy_pos++ ;
            /* (306 * R + 601 * G + 117 * B)/124 */
            mono = ( ((cp0>>3)&0x1f) * 306
                   + (((cp0<<2)&0x1c) + ((cp1>>6)&0x03)) * 601
                   + (cp1 & 0x1f) * 117 ) / 124 ;
            if( mono >= 256 ) {
                mono = 0 ;
            } else {
                mono = 255 - mono ;
            }
            data_buff[0][x] = (unsigned char)mono ;
        }
        //copy_pos = data_buff[1] ;
        copy_pos = (void *)(0xc0000000 + ((y+1) * 320 * 2)) ;
        for( x=0 ; x<320 ; x++ ) {
            cp1 = (unsigned short int)*copy_pos++ ;
            cp0 = (unsigned short int)*copy_pos++ ;
            /* (306 * R + 601 * G + 117 * B)/124 */
            mono = ( ((cp0>>3)&0x1f) * 306
                   + (((cp0<<2)&0x1c) + ((cp1>>6)&0x03)) * 601
                   + (cp1 & 0x1f) * 117 ) / 124 ;
            if( mono >= 256 ) {
                mono = 0 ;
            } else {
                mono = 255 - mono ;
            }
            data_buff[1][x] = (unsigned char)mono ;
        }

        /* extend */
        for( x=0,xx=0 ; x<320 ; x+=2,xx+=3 ) {
            data_exte[0][xx]   = data_buff[0][x] ;
            data_exte[0][xx+2] = data_buff[0][x+1] ;
            data_exte[2][xx]   = data_buff[1][x] ;
            data_exte[2][xx+2] = data_buff[1][x+1] ;
            data_exte[0][xx+1] = (data_buff[0][x]+data_buff[0][x+1]) / 2 ;
            data_exte[2][xx+1] = (data_buff[1][x]+data_buff[1][x+1]) / 2 ;
            data_exte[1][xx]   = (data_buff[0][x]+data_buff[1][x]) / 2 ;
            data_exte[1][xx+2] = (data_buff[0][x+1]+data_buff[1][x+1]) / 2 ;
            data_exte[1][xx+1] = (data_buff[0][x]+data_buff[0][x+1]+data_buff[1][x]+data_buff[1][x+1]) / 4 ;
        }

        for( i=0 ; i<3 ; i++ ) {

          surp_now  = yy & 0x01 ;
          surp_next = (yy+1) & 0x01 ;

          for( xx=0 ; xx<480 ; xx++ ) {
            surp = (unsigned short int)data_exte[i][xx] + data_surp[surp_now][xx] ;
            if( surp >= PRN_PSET_THRED4 ) {
                surp -= PRN_PSET_THRED4 ;
                prn_area_set( 370-yy, xx ) ;
            }
            data_surp[surp_now ][xx+1] += (surp * 7) / 16 ;
            data_surp[surp_next][xx+0] += (surp * 5) / 16 ;
            data_surp[surp_next][xx+1] += surp / 16 ;
            if( xx > 0 ) {
                data_surp[surp_next][xx-1] += (surp * 3) / 16 ;
            }
          }
          /* clear now line surp */
          for( xx=0 ; xx<=480 ; xx++ ) {
            data_surp[surp_now][xx] = 0 ;
          }

          yy++ ;
        }
    }

    prn_area_print() ;

    pnt = data ;
    *pnt++ = 0x0a ; /* LF */
    *pnt++ = 0x0a ; /* LF */
    *pnt++ = 0x0a ; /* LF */
    *pnt++ = 0x0a ; /* LF */
    *pnt++ = 0x0c ; /* FF */
    len = pnt - data ;
    (void)prn_send_multibyte( (unsigned char *)data, len ) ;
}

